require 'socket'
require 'timeout'

module RSpec
  class Statsd
    class << self
      #
      # @private
      #
      # Delegate everything to the instance
      #
      def method_missing(m, *args, &block)
        if instance.respond_to?(m)
          instance.public_send(m, *args, &block)
        else
          super
        end
      end
    end

    require 'singleton'
    include Singleton

    attr_reader :pid

    def initialize
      clone! unless File.exist?(cache_dir)
      configure!
      start!
    end

    def start!
      unless @pid
        @pid = Process.spawn("node #{cache_dir}/stats.js #{config}", out: '/dev/null')
        sleep(0.5)

        at_exit do
          if @pid
            Process.kill('INT', @pid)
            Process.waitpid2(@pid)
          end
        end
      end

      @pid
    end

    #
    # Query Statsd for some information and then read the response. This is
    # incredibly brittle and should be the first place you look when debugging
    # tests.
    #
    # @param [String] message
    #   the message to send
    #
    # @return [String]
    #   the response
    #
    def query(message)
      unless @connection
        @connection = TCPSocket.new(host, admin_port)
      end

      @connection.send("#{message}\\000", 0)

      # This is a really super hack way to make sure the socket actually returns
      # some information.
      response = []
      begin
        Timeout.timeout(0.1) do
          while line = @connection.gets
            response << line.strip
          end
        end
      rescue Timeout::Error
      end

      response.join("\n")
    end

    def host
      @host ||= '127.0.0.1'
    end

    def port
      @port ||= random_open_port
    end

    def admin_port
      @admin_port ||= random_open_port
    end

    private

    def random_open_port
      server = TCPServer.new('127.0.0.1', 0)
      port = server.addr[1]
      server.close

      port
    end

    def clone!
      %x|git clone https://github.com/etsy/statsd.git #{cache_dir} > /dev/null 2>&1|
    end

    def configure!
      File.open(config, 'w') do |f|
        f.write <<-EOH.gsub(/^ {10}/, '')
          {
            address: '#{host}',
            port:    #{port},

            mgmt_address: '#{host}',
            mgmt_port: #{admin_port}
          }
        EOH
      end
    end

    def config
      @config ||= File.join(cache_dir, 'config.js')
    end

    def cache_dir
      @cache_dir ||= File.expand_path('../../../tmp/statsd', __FILE__)
    end
  end
end
